from collective.watcherlist.interfaces import IWatcherList
from plone.namedfile import NamedBlobFile
from Products.CMFCore.utils import getToolByName
from Products.CMFFormController.FormAction import FormActionKey
from Products.CMFPlone.utils import base_hasattr
from Products.CMFPlone.utils import safe_unicode
from Products.Poi.adapters import IResponseContainer
from Products.Poi.utils import normalize_filename
from ZODB.POSException import ConflictError
from zope.annotation.interfaces import IAnnotations
from zope.schema._bootstrapinterfaces import ConstraintNotSatisfied

import logging
import transaction


logger = logging.getLogger("Poi")
PROFILE_ID = 'profile-Products.Poi:default'


def migrate_workflow_changes(context):
    """Migrate workflow changes from ids to titles.

    When a response changes the workflow state of an issue, this
    change is recorded in that response.  This used to be done by
    storing review state ids.  Currently this is done by storing
    review state titles.  Friendlier for the end user and translatable
    to boot.  This migration finds responses with review state ids in
    them and turns them into titles.
    """
    logger.info("Starting migration of workflow changes.")
    catalog = getToolByName(context, 'portal_catalog')
    wftool = getToolByName(context, 'portal_workflow')

    def get_state_title(state_id):
        # This neatly returns the input when there is no such review
        # state id, which happens when the 'state_id' is already a
        # title.
        return wftool.getTitleForStateOnType(state_id, 'PoiIssue')

    issue_brains = catalog.searchResults(portal_type='PoiIssue')
    logger.info("Found %s PoiIssues.", len(issue_brains))
    fixed = 0
    for brain in issue_brains:
        try:
            issue = brain.getObject()
        except (AttributeError, KeyError):
            logger.warn("AttributeError or KeyError getting tracker object at "
                        "%s", brain.getURL())
            continue
        folder = IResponseContainer(issue)
        made_changes = False
        for response in folder:
            for change in response.changes:
                # def add_change(self, id, name, before, after):
                if change['id'] != 'review_state':
                    continue
                before = get_state_title(change['before'])
                if change['before'] != before:
                    made_changes = True
                    change['before'] = before
                change['after'] = get_state_title(change['after'])
        if made_changes:
            fixed += 1
            if fixed % 100 == 0:
                logger.info("Committing transaction after fixing "
                            "%s PoiIssues; still busy... " % fixed)
                transaction.commit()
    logger.info("Migration completed.  %s PoiIssues needed fixing.", fixed)


def run_workflow_step(context):
    context.runImportStepFromProfile(PROFILE_ID, 'workflow')
    # Run the update security on the workflow tool.
    logger.info('Updating security settings.  This may take a while...')
    wf_tool = getToolByName(context, 'portal_workflow')
    wf_tool.updateRoleMappings()
    logger.info('Done updating security settings.')


def update_tracker_managers(context, testing=False):
    """Make sure all managers of a tracker get the TrackerManager role.

    We used to give users in the 'managers' field of a tracker the
    local Manager role.  Now we give them the new TrackerManager role.
    In this upgrade step we remove the local Manager role from all
    users.  Instead we give them the TrackerManager role.  This might
    give a few 'false positives', where a user has been intentionally
    given the local Manager role and will now lose it.  That can't be
    helped and is not expected to be a big problem.

    If testing is True, do not commit; this avoids some problems when
    running the tests.
    """
    logger.info("Starting update of tracker managers.")
    catalog = getToolByName(context, 'portal_catalog')
    tracker_brains = catalog.searchResults(
        portal_type=('PoiTracker', 'PoiPscTracker'))
    logger.info("Found %s PoiTrackers.", len(tracker_brains))
    for brain in tracker_brains:
        try:
            tracker = brain.getObject()
        except (AttributeError, KeyError):
            logger.warn("AttributeError or KeyError getting tracker object at "
                        "%s", brain.getURL())
            continue

        field = tracker.getField('managers')
        managers = field.get(tracker)
        tracker_changed = False
        for user_id in managers:
            user_changed = False
            local_roles = list(tracker.get_local_roles_for_userid(user_id))
            if 'Manager' in local_roles:
                logger.info("Removing local Manager role from user %s.",
                            user_id)
                local_roles.remove('Manager')
                user_changed = True
            if 'TrackerManager' not in local_roles:
                logger.info("Giving user %s TrackerManager role.", user_id)
                local_roles.append('TrackerManager')
                user_changed = True
            if user_changed:
                tracker.manage_setLocalRoles(user_id, local_roles)
                tracker_changed = True
        if tracker_changed and not testing:
            logger.info("Committing after updating roles on tracker %s",
                        tracker.absolute_url())
            transaction.commit()


def remove_form_controller_action(context):
    """Remove our action from the portal_form_controller.

    We used to add the poi_issue_post action (a skin script) to the
    validate_integrity script.  Now we have registered an event
    handler instead so the action is no longer needed.  And we have
    removed the skin script, so we must remove the action as otherwise
    anyone submitting a new issue will get an error.
    """
    controller = getToolByName(context, 'portal_form_controller')
    action_key = FormActionKey('validate_integrity', 'success', 'PoiIssue', '')
    try:
        action = controller.actions.get(action_key)
    except KeyError:
        logger.info("Action for validate_integrity not found; ignoring.")
        return
    if action.getActionArg() != 'string:poi_issue_post':
        logger.info("Expected action argument 'string:poi_issue_post' for "
                    "successfull validate_integrity on PoiIssue; found %r "
                    "instead; ignoring.", action.getActionArg())
        return
    try:
        controller.actions.delete(action_key)
    except (ConflictError, KeyboardInterrupt):
        raise
    except Exception, exc:
        # This way of removing an action is yucky; I wonder how many
        # things can go wrong; let's inform the user.
        url = controller.absolute_url() + '/manage_formActionsForm'
        error = ("Exception while removing action poi_issue_post on "
                 "validate_integrity. Please go to %s and remove it manually. "
                 "Original exception was: %r" % (url, exc))
        raise Exception(error)
    else:
        logger.info("Removed action poi_issue_post from validate_integrity.")


def run_rolemap_step(context):
    context.runImportStepFromProfile(PROFILE_ID, 'rolemap')


def run_sharing_step(context):
    context.runImportStepFromProfile(PROFILE_ID, 'sharing')


def run_types_step(context):
    context.runImportStepFromProfile(PROFILE_ID, 'typeinfo')


def purge_workflow_scripts(context):
    """Remove reference to old workflow scripts.

    We used to have some external methods hooked up to the issue
    workflow, but this is now done with events.  Having the scripts
    mentioned in the ZMI does not seem to cause problems, but it is
    cleaner to remove them, as we do not need them.  Also, a workflow
    export and import would fail.
    """
    wf_tool = getToolByName(context, 'portal_workflow')
    wf_id = 'poi_issue_workflow'
    bad_scripts = ('sendInitialEmail', 'sendResolvedMail')
    wf = wf_tool.getWorkflowById(wf_id)
    for script_name in bad_scripts:
        if script_name in wf.scripts.objectIds():
            wf.scripts._delObject(script_name)
            logger.info('Removed script %s from %s', script_name, wf_id)


def run_javascript_step(context):
    context.runImportStepFromProfile(PROFILE_ID, 'jsregistry')


def migrate_tracker_watchers(context):
    """Migrate tracker watchers.

    Watchers of a tracker were first stored in annotations, but should
    now be stored in a LinesField.
    """
    logger.info("Starting update of tracker watchers.")
    catalog = getToolByName(context, 'portal_catalog')
    tracker_brains = catalog.searchResults(
        portal_type=('PoiTracker', 'PoiPscTracker'))
    logger.info("Found %s PoiTrackers.", len(tracker_brains))
    for brain in tracker_brains:
        try:
            tracker = brain.getObject()
        except (AttributeError, KeyError):
            logger.warn("AttributeError or KeyError getting tracker object at "
                        "%s", brain.getURL())
            continue

        watchers = IWatcherList(tracker)
        # We would want to check watchers.__mapping, but that fails.
        # watchers._WatcherList__mapping would work, but it looks
        # suspicious to me.  So let's get the annotations.
        annotations = IAnnotations(tracker)
        mapping = annotations.get(watchers.ANNO_KEY, None)
        if not mapping:
            continue
        try:
            old_value = mapping['watchers']
        except KeyError:
            continue
        logger.info("Setting watchers of tracker at %s: %r",
                    tracker.absolute_url(), old_value)
        tracker.setWatchers(old_value)
        del mapping['watchers']


def recook_resources(context):
    context.portal_javascripts.cookResources()
    context.portal_css.cookResources()


def migrate_response_attachments_to_blobstorage(context):
    logger.info('Migrating response attachments to blob storage.')
    catalog = getToolByName(context, 'portal_catalog')
    already_migrated = 0
    migrated = 0
    for brain in catalog.unrestrictedSearchResults(portal_type='PoiIssue'):
        path = brain.getPath()
        try:
            issue = brain.getObject()
        except (AttributeError, ValueError, TypeError):
            logger.warn('Error getting object from catalog for path %s', path)
            continue
        folder = IResponseContainer(issue)
        for id, response in enumerate(folder):
            if response is None:
                # Has been removed.
                continue
            attachment = response.attachment
            if attachment is None:
                continue
            if isinstance(attachment, NamedBlobFile):
                # Already migrated
                logger.debug('Response %d already migrated, at %s.', id, path)
                already_migrated += 1
                continue
            content_type = getattr(attachment, 'content_type', '')
            filename = getattr(attachment, 'filename', '')
            if not filename and hasattr(attachment, 'getId'):
                filename = attachment.getId()
            data = attachment.data
            # Data can be 'nested' in OFS.Image.Pdata.
            if base_hasattr(data, 'data'):
                data = data.data
            filename = safe_unicode(filename)
            try:
                blob = NamedBlobFile(
                    data, contentType=content_type, filename=filename)
            except ConstraintNotSatisfied:
                # Found in live data: a filename that includes a newline...
                logger.info('Trying to normalize filename %s', filename)
                filename = normalize_filename(filename, context.REQUEST)
                logger.info('Normalize to %s', filename)
                blob = NamedBlobFile(
                    data, contentType=content_type, filename=filename)
            response.attachment = blob
            logger.debug('Response %d migrated, at %s.', id, path)
            migrated += 1

    logger.info('Migrated %d response attachments to blobs. '
                '%d already migrated.', migrated, already_migrated)


def migrate_issue_attachments_to_blobstorage(context):
    from plone.app.blob.migrations import migrate
    logger.info('Migrating to blob attachments for issues. '
                'This can take a long time...')
    # Technically, the plone.app.blob migration says it is only for non blob
    # fields that are still in the schema but are overridden using
    # archetypes.schemaextender with a blob field with the same name.  But it
    # seems to go fine for Issues where we have simply changed the schema
    # directly.  The getters of a BlobFileField and normal FileField don't
    # differ that much.
    migrate(context, 'PoiIssue')
    logger.info("Done migrating to blob attachment for issues.")



from Acquisition import aq_inner
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from Products.PythonScripts.standard import url_quote
from ZTUtils import make_query

ACTIVE_STATES = ['open', 'in-progress', 'unconfirmed']


class IssueFolderView(BrowserView):

    issue_portal_type = 'PoiIssue'

    def getActiveStates(self):
        return ACTIVE_STATES

    def getFilteredIssues(self, criteria=None, **kwargs):
        """Get the contained issues in the given criteria.
        """
        context = aq_inner(self.context)
        query = self.buildIssueSearchQuery(criteria, **kwargs)
        catalog = getToolByName(context, 'portal_catalog')
        return catalog.searchResults(query)

    def getIssueSearchQueryString(self, criteria=None, **kwargs):
        """Return a query string for an issue query.

        Form of return string:name1=value1&name2=value2
        """
        query = self.buildIssueSearchQuery(criteria, **kwargs)
        return make_query(query)

    def buildIssueSearchQuery(self, criteria=None, **kwargs):
        """Build canonical query for issue search.
        """
        context = aq_inner(self.context)

        if criteria is None:
            criteria = kwargs
        else:
            criteria = dict(criteria)

        allowedCriteria = {
            'release': 'getRelease',
            'area': 'getArea',
            'issueType': 'getIssueType',
            'severity': 'getSeverity',
            'targetRelease': 'getTargetRelease',
            'state': 'review_state',
            'tags': 'Subject',
            'responsible': 'getResponsibleManager',
            'creator': 'Creator',
            'text': 'SearchableText',
            'id': 'getId',
        }

        query = {}
        query['path'] = '/'.join(context.getPhysicalPath())
        query['portal_type'] = [self.issue_portal_type]

        for k, v in allowedCriteria.items():
            if criteria.get(k):
                query[v] = criteria[k]
            elif criteria.get(v):
                query[v] = criteria[v]

        # Playing nicely with the form.

        # Subject can be a string of one tag, a tuple of several tags
        # or a dict with a required query and an optional operator
        # 'and/or'.  We must avoid the case of the dict with only the
        # operator and no actual query, else we will suffer from
        # KeyErrors.  Actually, when coming from the
        # poi_issue_search_form, instead of say from a test, its type
        # is not 'dict', but 'instance', even though it looks like a
        # dict.  See http://plone.org/products/poi/issues/137
        if 'Subject' in query:
            subject = query['Subject']
            # We cannot use "subject.has_key('operator')" or
            # "'operator' in subject'" because of the strange
            # instance.
            try:
                subject['operator']
            except TypeError:
                # Fine: subject is a string or tuple.
                pass
            except KeyError:
                # No operator, so nothing can go wrong.
                pass
            else:
                try:
                    subject['query']
                except KeyError:
                    del query['Subject']

        query['sort_on'] = criteria.get('sort_on', 'created')
        query['sort_order'] = criteria.get('sort_order', 'reverse')
        if criteria.get('sort_limit'):
            query['sort_limit'] = criteria.get('sort_limit')

        return query

    def getMyIssues(self, openStates=['open', 'in-progress'],
                    memberId=None, manager=False):
        """Get a catalog query result set of my issues.

        So: all issues assigned to or submitted by the current user,
        with review state in openStates.

        If manager is True, add unconfirmed to the states.
        """
        context = aq_inner(self.context)
        if not memberId:
            mtool = getToolByName(context, 'portal_membership')
            member = mtool.getAuthenticatedMember()
            memberId = member.getId()

        if manager:
            if 'unconfirmed' not in openStates:
                openStates += ['unconfirmed']

        issues = []

        for i in self.getFilteredIssues(state=openStates):
            responsible = i.getResponsibleManager
            creator = i.Creator
            if memberId in (creator, responsible) or \
                    (manager and responsible == '(UNASSIGNED)'):
                issues.append(i)

        return issues

    def getOrphanedIssues(self, openStates=['open', 'in-progress'],
                          memberId=None):
        """Get a catalog query result set of orphaned issues.

        Meaning: all open issues not assigned to anyone and not owned
        by the given user.
        """
        context = aq_inner(self.context)
        if not memberId:
            mtool = getToolByName(context, 'portal_membership')
            member = mtool.getAuthenticatedMember()
            memberId = member.getId()

        issues = []

        for i in self.getFilteredIssues(state=openStates):
            responsible = i.getResponsibleManager
            creator = i.Creator
            if creator != memberId and responsible == '(UNASSIGNED)':
                issues.append(i)

        return issues


class QuickSearchView(BrowserView):
    """Parse a quicksearch string and jump to the appropriate issue or
    search result page.

    """

    def __call__(self):
        tracker = aq_inner(self.context)
        search_text = self.request.form.get('searchText', '')
        issue_id = search_text
        if issue_id.startswith('#'):
            issue_id = issue_id[1:]
        base_url = tracker.absolute_url()
        if issue_id in tracker.keys():
            url = '%s/%s' % (base_url, issue_id)
        else:
            url = '%s/poi_issue_search?SearchableText=%s' % (
                base_url, url_quote(search_text))
        self.request.RESPONSE.redirect(url)



from setuptools import setup, find_packages

readmefile = open('README.rst')
readme = readmefile.read().strip()
readmefile.close()

historyfile = open('CHANGES.rst')
history = historyfile.read().strip()
historyfile.close()

long_description = readme + "\n\n" + history

setup(name='Products.Poi',
      version='2.3',
      description="Poi: A friendly issue tracker",
      long_description=long_description,
      # Get more strings from
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
          "Framework :: Plone",
          "Framework :: Plone :: 4.1",
          "Framework :: Plone :: 4.2",
          "Framework :: Plone :: 4.3",
          "Framework :: Zope2",
          "Framework :: Zope3",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.6",
          "Programming Language :: Python :: 2.7",
          "Topic :: Software Development :: Libraries :: Python Modules",
          ],
      keywords='bugs issue tracker',
      author='Martin Aspeli',
      author_email='plone-users@lists.sourceforge.net',
      url='http://plone.org/products/poi',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['Products'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'Products.CMFPlone>=4.0b1',
          'Products.AddRemoveWidget>=1.4.2',
          'Products.DataGridField>=1.9.2',
          'collective.watcherlist>=0.2',
          'plone.app.blob',
          'plone.namedfile',
      ],
      extras_require={
          'test': [
              'Products.PloneTestCase',
          ],
      },
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



# Permissions used by Poi

from Products.CMFCore import permissions as CMFCorePermissions

AccessContentsInformation = CMFCorePermissions.AccessContentsInformation
ModifyPortalContent = CMFCorePermissions.ModifyPortalContent
View = CMFCorePermissions.View

EditResponse = "Poi: Edit response"
ManageTracker = "Poi: Manage tracker"
ModifyIssueAssignment = "Poi: Modify issue assignment"
ModifyIssueSeverity = "Poi: Modify issue severity"
ModifyIssueState = "Poi: Modify issue state"
ModifyIssueTags = "Poi: Modify issue tags"
ModifyIssueTargetRelease = "Poi: Modify issue target release"
ModifyIssueWatchers = "Poi: Modify issue watchers"
UploadAttachment = "Poi: Upload attachment"



from Testing import ZopeTestCase

from Products.Poi.tests import ptc

default_user = ZopeTestCase.user_name


class TestFeeds(ptc.PoiTestCase):
    """Test RSS feed functionality"""

    def afterSetUp(self):
        self.workflow = self.portal.portal_workflow
        self.membership = self.portal.portal_membership
        self.addMember('member1', 'Member One', 'member1@example.com',
                       ['Member'], '2005-01-01')
        self.addMember('member2', 'Member Two', 'member2@example.com',
                       ['Member'], '2005-01-01')
        self.addMember('member3', 'Member Three', 'member3@example.com',
                       ['Member'], '2005-01-01')
        self.tracker = self.createTracker(
            self.folder, 'issue-tracker',
            managers=('member1', 'member2', default_user))
        self.issues = []

    def addIssue(self, title, responsible='(UNASSIGNED)',
                 transition='accept-unconfirmed'):
        issue = self.createIssue(self.tracker, title,
                                 responsibleManager=responsible)
        userId = self.membership.getAuthenticatedMember().getId()
        self.login(self.tracker.getManagers()[0])
        self.workflow.doActionFor(issue, transition)
        self.login(userId)
        issue.reindexObject()
        self.issues.append(issue)

    def testGetMyIssues(self):
        # Creator = default_user
        # 1: owned by default_user, assigned to member1
        self.addIssue('A:member1', 'member1')
        # 2: owned by default_user, not assigned to member1
        self.addIssue('A:member2', 'member2')

        # Creator = member1
        self.login('member1')
        # 3: owned by member1, not assigned to anyone
        self.addIssue('C:member1')
        # 4: owned by member1, assigned to default_user
        self.addIssue('A:default', default_user)

        # Creator = member 3 (not in tracker)
        self.login('member3')
        # 5: owned by member3, not assigned to anyone
        self.addIssue('C:member3')

        self.login(default_user)

        # Wrong state
        # 6: owned by and assigned to default_user, rejected
        self.addIssue('S:rejected', default_user, 'reject-unconfirmed')

        issuefolder = self.tracker.restrictedTraverse('@@issuefolder')
        myIssues = issuefolder.getMyIssues(memberId=default_user)
        ids = sorted([int(i.getId) for i in myIssues])
        self.assertEqual([1, 2, 4], ids)

        myIssues = issuefolder.getMyIssues(memberId='member1')
        ids = [int(i.getId) for i in myIssues]
        ids.sort()
        self.assertEqual([1, 3, 4], ids)

        myIssues = issuefolder.getMyIssues(openStates=['closed'],
                                           memberId=default_user)
        self.assertEqual(len(myIssues), 0)

        myIssues = issuefolder.getMyIssues(memberId='member3')
        self.assertEqual(len(myIssues), 1)
        self.assertEqual(myIssues[0].getId, '5')

        myIssues = issuefolder.getMyIssues(openStates=['rejected'])
        self.assertEqual(len(myIssues), 1)
        self.assertEqual(myIssues[0].getId, '6')

    def testGetOrphanedIssues(self):
        # Creator = default_user
        # 1: owned by default_user, assigned to member1
        self.addIssue('A:member1', 'member1')
        # 2: owned by default_user, not assigned to member1
        self.addIssue('A:member2', 'member2')

        # Creator = member1
        self.login('member1')
        # 3: owned by member1, not assigned to anyone
        self.addIssue('C:member1')
        # 4: owned by member1, assigned to default_user
        self.addIssue('A:default', default_user)

        # Creator = member 3 (not in tracker)
        self.login('member3')
        # 5: owned by member3, not assigned to anyone
        self.addIssue('C:member3')

        self.login(default_user)

        # Wrong state
        # 6: owned by and assigned to default_user, rejected
        self.addIssue('S:rejected', default_user, 'reject-unconfirmed')

        issuefolder = self.tracker.restrictedTraverse('@@issuefolder')
        myIssues = issuefolder.getOrphanedIssues(memberId=default_user)
        ids = sorted([int(i.getId) for i in myIssues])
        self.assertEqual([3, 5], ids)

        myIssues = issuefolder.getOrphanedIssues(memberId='member1')
        ids = [int(i.getId) for i in myIssues]
        ids.sort()
        self.assertEqual([5], ids)

        myIssues = issuefolder.getOrphanedIssues(openStates=['closed'],
                                                 memberId=default_user)
        self.assertEqual(len(myIssues), 0)

        myIssues = issuefolder.getOrphanedIssues(memberId='member3')
        self.assertEqual(len(myIssues), 1)
        self.assertEqual(myIssues[0].getId, '3')


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(TestFeeds))
    return suite
